# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from . import base


class OdooAccount(base.OdooModule):
    _name = "Account"
    _key = "account"

    def apply(self):
        super(OdooAccount, self).apply()

        first = True
        banks = self._datas.get(self._key, {}).get('data', {}).get('banks', [])
        for bank in banks:
            # Create or update bank
            vals = {
                'name': bank,
                'bic': banks.get(bank).get("bic", ""),
                'street': banks.get(bank).get("street", ""),
                'street2': banks.get(bank).get("street2", ""),
                'city': banks.get(bank).get("city", ""),
                'zip': banks.get(bank).get("zip", ""),
                'country': self.execute_odoo('res.country', 'search',
                                             [[('code', '=', banks.get(bank)["country_code"])], 0, 0, "id", False],
                                             {'context': self._context})[0],
                'phone': banks.get(bank).get("phone", ""),
                'email': banks.get(bank).get("email", ""),
            }
            bank_id = self.execute_odoo('res.bank', 'search', [[('bic', '=', vals["bic"])], 0, 0, "id", False],
                                        {'context': self._context})
            if not bank_id:
                bank_id = self.execute_odoo('res.bank', 'search', [[('name', '=', vals["name"])], 0, 0, "id", False],
                                            {'context': self._context})
                if not bank_id:
                    bank_id = self.execute_odoo('res.bank', 'create', [vals], {'context': self._context})
                else:
                    self.execute_odoo('res.bank', 'write', [bank_id, vals], {'context': self._context})
                    bank_id = bank_id[0]
            else:
                self.execute_odoo('res.bank', 'write', [bank_id, vals], {'context': self._context})
                bank_id = bank_id[0]

            # Create or update partner bank
            vals = {
                'acc_number': banks.get(bank).get("iban", ""),
                'partner_id': 1,
                'bank_id': bank_id,
            }
            partner_bank_id = self.execute_odoo('res.partner.bank', 'search',
                                                [[('acc_number', '=', vals["acc_number"])], 0, 0, "id", False],
                                                {'context': self._context})
            if not partner_bank_id:
                partner_bank_id = self.execute_odoo('res.partner.bank', 'create', [vals], {'context': self._context})
            else:
                self.execute_odoo('res.partner.bank', 'write', [partner_bank_id, vals], {'context': self._context})
                partner_bank_id = partner_bank_id[0]

            # Check default journal exist
            default_journal_id = self.execute_odoo('account.journal', 'search',
                                                   [[('name', '=', 'Banque')], 0, 0, "id", False],
                                                   {'context': self._context})
            journal_id = False
            if first:
                first = False
                if default_journal_id:
                    journal_id = default_journal_id

            # Create or update journal
            vals = {
                'name': banks.get(bank).get("journal_name", ""),
                'code': banks.get(bank).get("code", ""),
                'bank_account_id': partner_bank_id,
                'type': 'bank',
            }
            if not journal_id:
                journal_id = self.execute_odoo('account.journal', 'search',
                                               [[('bank_account_id', '=', vals['bank_account_id'])], 0, 0, "id", False],
                                               {'context': self._context})
            if not len(journal_id):
                journal_id = self.execute_odoo('account.journal', 'create', [vals], {'context': self._context})
            else:
                self.execute_odoo('account.journal', 'write', [journal_id, vals], {'context': self._context})
                journal_id = journal_id[0]

            # Update account
            account_id = self.execute_odoo('account.journal', 'read', [journal_id, ['default_debit_account_id']],
                                           {'context': self._context})[0]['default_debit_account_id'][0]
            self.execute_odoo('account.account', 'write', [account_id, {'name': vals["name"]}],
                              {'context': self._context})

        # # ##########################################################################################################
        # # JOURNAL
        # # ##########################################################################################################
        # sale_journal_id = self.execute_odoo('account.journal', 'search',
        #                                     [[('name', '=', "Factures clients")], 0, 0, "id", False],
        #                                     {'context': self._context})
        # sale_journal = \
        # self.execute_odoo('account.journal', 'read', [sale_journal_id, ['sequence_id', 'refund_sequence_id']],
        #                   {'context': self._context})[0]
        # if config.get(key, {}).get('data', {}).get('invoice', ""):
        #     self.execute_odoo('ir.sequence', 'write', [sale_journal['sequence_id'][0], {
        #         'prefix': config.get(key, {}).get('data', {}).get('invoice', "")}], {'context': context})
        # if config.get(key, {}).get('data', {}).get('refund', ""):
        #     self.execute_odoo('ir.sequence', 'write', [sale_journal['refund_sequence_id'][0], {
        #         'prefix': config.get(key, {}).get('data', {}).get('refund', "")}], {'context': context})
        #
        # purchase_journal_id = self.execute_odoo('account.journal', 'search',
        #                                         [[('name', '=', "Factures fournisseurs")], 0, 0, "id", False],
        #                                         {'context': self._context})
        # purchase_journal = \
        # self.execute_odoo('account.journal', 'read', [purchase_journal_id, ['sequence_id', 'refund_sequence_id']],
        #                   {'context': self._context})[0]
        # if config.get(key, {}).get('data', {}).get('supplier_invoice', ""):
        #     self.execute_odoo('ir.sequence', 'write', [purchase_journal['sequence_id'][0], {
        #         'prefix': config.get(key, {}).get('data', {}).get('supplier_invoice', "")}], {'context': context})
        # if config.get(key, {}).get('data', {}).get('supplier_refund', ""):
        #     self.execute_odoo('ir.sequence', 'write', [purchase_journal['refund_sequence_id'][0], {
        #         'prefix': config.get(key, {}).get('data', {}).get('supplier_refund', "")}], {'context': context})



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from . import connection  # noqa
from . import modules  # noqa
from . import translations  # noqa
from . import datas  # noqa
from . import config  # noqa
from . import roles  # noqa
from . import defaults  # noqa
from . import users  # noqa
from . import account  # noqa
from . import website  # noqa
from . import imports  # noqa
from . import import_configurator  # noqa
from . import call  # noqa



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import logging
from collections import OrderedDict

from ..logging import get_logger


class OdooModule:
    # pylint: disable=too-many-instance-attributes
    def __init__(self, configurator):
        self._configurator = configurator
        self._connection = configurator.connection
        self._password_manager = configurator.password_manager
        self._import_manager = configurator.import_manager
        self._datas = configurator.config
        self._pre_datas = configurator.pre_update_config
        self._mode = configurator.mode
        self._debug = configurator.debug
        if self._connection:
            self._context = self._connection.context.copy()
            self.execute_odoo = self._connection.execute_odoo
        self.logger = get_logger(self._name.ljust(15))
        if self._debug:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)
        self.apply()

    @property
    def _name(self):
        raise NotImplementedError

    def _published_objects(self):
        return self._connection, self._password_manager

    def install_mode(self):
        if 'install' in self._mode:
            return True
        return False

    def get_mapping_method(self):
        res = {}
        for o in self._published_objects():
            res[o.__class__] = [method for method in dir(o.__class__) if method.startswith('_') is False]
        return res

    def deep_convert_dict(self, layer):
        to_ret = layer
        if isinstance(layer, OrderedDict):
            to_ret = dict(layer)
        try:
            for key, value in to_ret.items():
                to_ret[key] = self.deep_convert_dict(value)
        except AttributeError:
            pass
        return to_ret

    def safe_eval(self, name):
        for object_class, methods in self.get_mapping_method().items():
            if name.split('(')[0] in methods:
                for o in self._published_objects():
                    if o.__class__ == object_class:
                        self.logger.debug("o.%s", name)
                        try:
                            return eval("o.%s" % name)
                        except Exception as err:
                            raise err
        raise Exception("Cannot eval %s" % name)

    def pre_config(self, config, rec=0):
        if config and isinstance(config, dict) or isinstance(config, OrderedDict):
            for key in config:
                if isinstance(config[key], list):
                    tmp = []
                    for i in config[key]:
                        if type(i) is str and i.startswith('get_'):
                            i = self.safe_eval(i)
                        tmp.append(i)
                    config[key] = tmp
                elif isinstance(config[key], str) and config[key].startswith('get_'):
                    config[key] = self.safe_eval(config[key])
                elif isinstance(config[key], (dict, OrderedDict)) and config[key].get(
                        'values', '') and not config[key].get('force_id', ''):
                    model = config[key].get('model')
                    if config[key].get('values'):
                        self.pre_config(config[key]['values'], rec + 1)
                    values = config[key].get('values')
                    search_key = config[key].get('key', 'name')
                    domain = []
                    for i in search_key.split(','):
                        if values.get(i, False):
                            domain.append((i, '=', values[i]))
                    if config[key].get('update_domain', False):
                        continue

                    object_ids = self.execute_odoo(model, 'search', [domain, 0, 0, "id", False],
                                                   {'context': self._context})
                    self.logger.debug("%s %s %s", model, domain, object_ids)
                    if object_ids:
                        config[key] = object_ids[0]
                        if values:
                            self.execute_odoo(model, 'write', [config[key], dict(values)], {'context': self._context})
                    else:
                        config[key] = self.execute_odoo(model, 'create', [dict(values)], {'context': self._context})
                elif isinstance(config[key], (dict, OrderedDict, list)):
                    if not self.install_mode() and config[key].get('on_install_only', False):
                        continue
                    for i in config[key]:
                        key_value = config[key][i]
                        if isinstance(key_value, str) and key_value.startswith('get_'):
                            config[key][i] = self.safe_eval(key_value)
                        self.pre_config(config[key][i], rec + 1)

    def apply(self):
        self.logger.info("Apply %s", self._name)



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import os
from .logging import get_logger
from pykeepass import PyKeePass
import cryptocode

logger = get_logger(__name__)
CACHE = "/tmp/.configurator_cache"


class PasswordManager:

    def __init__(self, password):
        self._keepass_password = password

    def get_keepass(self, filename):
        return PyKeePass(os.getenv("HOME") + "/" + filename, password=self._keepass_password)

    def get_keepass_entry(self, filename, group, entry):
        kp = self.get_keepass(filename)
        kgroup = kp.find_groups(name=group, first=True)
        if kgroup:
            kentry = kp.find_entries(title=entry, group=kgroup, recursive=False, first=True)
            if kentry:
                return kentry
            else:
                logger.error("Cannot find %s in %s" % (entry, group))
                raise Exception("Cannot find %s in %s" % (entry, group))
        else:
            logger.error("Cannot find %s" % group)
            raise Exception("Cannot find %s" % group)
        return False

    def get_keepass_entry_value(self, filename, group, entry, field, default='admin'):
        try:
            entry = self.get_keepass_entry(filename, group, entry)
        except Exception as err:
            logger.error(type(err).__name__)
            return False
        if entry:
            return eval('entry.' + field)
        return default

    def get_keepass_password(self, filename, group, entry):
        return self.get_keepass_entry_value(filename, group, entry, 'password')

    def get_keepass_password_crypt(self, filename, group, entry, key):
        return cryptocode.encrypt(self.get_keepass_entry_value(filename, group, entry, 'password'), key)

    def get_keepass_user(self, filename, group, entry):
        return self.get_keepass_entry_value(filename, group, entry, 'username')

    def get_keepass_url(self, filename, group, entry):
        kp_url = self.get_keepass_entry_value(filename, group, entry, 'url')
        if kp_url and kp_url[-1] == '/':
            kp_url = kp_url[:-1]
        return kp_url



#!/usr/bin/env python3
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from src.odoo_configurator.__main__ import main

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from . import base


class OdooUsers(base.OdooModule):
    _name = "Users"
    _key = "users"

    def apply(self):
        super(OdooUsers, self).apply()

        users = self._datas.get(self._key, {}).get('datas', {})
        for user in users:
            groups_id = []
            for group in users[user].get("groups_id", []):
                if group == "unlink all":
                    groups_id.append((5,))
                else:
                    groups_id.append(
                        (4, self._connection.get_ref(group)))

            vals = {
                'login': users[user]['values'].get("login"),
                'groups_id': groups_id,
            }
            self._context['active_test'] = False
            user_id = self.execute_odoo('res.users', 'search', [[('login', '=', vals['login'])], 0, 0, "id", False],
                                        {'context': self._context})
            if not user_id:
                self.execute_odoo('res.users', 'create', [vals], {'context': self._context})
            else:
                self.execute_odoo('res.users', 'write', [user_id, vals], {'context': self._context})



#!/usr/bin/env python3
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from .configurator import Configurator
import argparse


def main():
    parser = argparse.ArgumentParser(description='Odoo Configurator')
    parser.add_argument('files', metavar='files', type=open, nargs='+', help='Config Files To load')
    parser.add_argument('--update', action='store_true', help='Update Mode')
    parser.add_argument('--install', action='store_true', help='Install Mode')
    parser.add_argument('--debug', action='store_true', help='Debug log')
    parser.add_argument('--debug_xmlrpc', action='store_true', help='Debug log xmlrpc')
    parser.add_argument('--keepass', type=str, help='Keepass password')
    args = parser.parse_args()

    c = Configurator([i.name for i in args.files], install=args.install, update=args.update, debug=args.debug,
                     debug_xmlrpc=args.debug_xmlrpc, keepass=args.keepass)
    c.show()
    log = c.start()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from . import base
from .modules import OdooModules


class OdooWebsite(base.OdooModule):
    _name = "Website"
    _key = "website"

    def apply(self):
        super(OdooWebsite, self).apply()

        theme = self._datas.get(self._key, {}).get("theme", False)
        if theme:
            OdooModules(self._configurator).install_odoo_theme(theme)



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from . import base


class OdooTranslations(base.OdooModule):
    _name = "Translations"
    _key = "translations"

    def apply(self):
        super(OdooTranslations, self).apply()
        self._context['active_test'] = False
        translations = self._datas.get(self._key, {})
        installed_translations = []
        odoo_lang = self.execute_odoo('res.lang', 'search_read', [[('code', 'in', translations)], ['id', 'code', 'active']],
                                      {'context': self._context})
        for lang_code in translations:
            for lang in odoo_lang:
                if lang['code'] == lang_code:
                    if lang['active']:
                        self.logger.info("\t- Translation %s already installed" % lang_code)
                    else:
                        self.logger.info("\t- Install %s Translation" % lang_code)
                        if self._connection._version >= 16:
                            wizard_id = self.execute_odoo('base.language.install', 'create',
                                                          [{'lang_ids': [lang['id']]}])
                        else:
                            wizard_id = self.execute_odoo('base.language.install', 'create', [{'lang': lang_code}])
                        self.execute_odoo('base.language.install', 'lang_install', [wizard_id])
                    installed_translations.append(lang_code)
        missing_translations = set(translations) - set(installed_translations)
        if missing_translations:
            raise Exception("Translations not found : %s" % ", ".join(missing_translations))



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import logging


def get_logger(name):
    logger = logging.getLogger(name)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(name)s : %(message)s')
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.handlers.clear()
    logger.addHandler(ch)
    logger.setLevel(logging.INFO)
    return logger



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from collections import OrderedDict
from . import base


class OdooConfig(base.OdooModule):
    _name = "Config"

    def apply(self):
        super(OdooConfig, self).apply()
        config_values = self.prepare_config_values()
        if config_values:
            self.execute_config(config_values)

    def execute_script_config(self, datas=False):
        datas = datas if datas else self._datas
        for key in datas:
            if isinstance(datas.get(key), dict) or isinstance(datas.get(key), OrderedDict):
                if datas.get(key).get('config', {}):
                    values = self.prepare_company_values(datas.get(key).get('config', {}))
                    self.execute_config(values)

    def prepare_config_values(self, datas=False):
        datas = datas if datas else self._datas
        config_values = {}
        for key in datas:
            if isinstance(datas.get(key), dict) or isinstance(datas.get(key), OrderedDict):
                if datas.get(key).get('config', {}):
                    self.logger.info("\t- Config for %s" % key)
                    config_values.update(datas.get(key).get('config', {}))
        if config_values:
            self.prepare_company_values(datas)
        return config_values

    def prepare_company_values(self, values):
        self.pre_config(values)
        if values.get('company_id'):
            self._context['allowed_company_ids'] = [values.get('company_id')]
        if values.get('allowed_company_ids'):
            self._context['allowed_company_ids'] = values.get('allowed_company_ids')
            values.pop('allowed_company_ids')
        return values

    def execute_config(self, config):
        for key in config:
            if isinstance(config[key], str) and config[key].startswith('get_'):
                config[key] = self.safe_eval(config[key])
        domain = []
        config_ids = self.execute_odoo('res.config.settings', 'search', [domain], {'context': self._context})
        if config_ids:
            config_id = config_ids[-1]
        else:
            config_id = self.execute_odoo('res.config.settings', 'create', [self.deep_convert_dict(config)],
                                          {'context': self._context})
        self.execute_odoo('res.config.settings', 'write', [config_id, self.deep_convert_dict(config)],
                          {'context': self._context})
        self.execute_odoo('res.config.settings', 'execute', [config_id], {'context': self._context})



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from collections import OrderedDict
import os
import sys
from . import base


class OdooImports(base.OdooModule):
    _name = "Imports"

    def get_func(self, data):
        func = getattr(self._import_manager, 'import_csv')
        if data.get('specific_import', False):
            import importlib.util as ilu
            self.logger.info("Import specific import %s %s : %s" % (
                data.get('specific_import', False), data.get('specific_method', False), data.get('file_path', False)))
            path = os.path.join(os.path.dirname(sys.argv[1]), 'datas', data.get('specific_import', False))
            spec = ilu.spec_from_file_location('import_specific', path)
            specific_lib = ilu.module_from_spec(spec)
            spec.loader.exec_module(specific_lib)
            func = getattr(self._import_manager, data.get('specific_method', False))
        return func

    def apply(self):
        super(OdooImports, self).apply()
        datas = self._datas.get('import_data', {})
        if datas:
            self.logger.info("\t- %s" % "Global")
        for import_name in datas:
            if not self.install_mode() and datas[import_name].get('on_install_only', False):
                return
            import_data = datas[import_name]
            func = self.get_func(import_data)
            func(import_data.get('file_path', ''), import_data.get('model', ''), params=import_data)
        for key in self._datas:
            if isinstance(self._datas.get(key), dict) or isinstance(self._datas.get(key), OrderedDict):
                for key_import, import_data in self._datas.get(key, {}).get('import_data', {}).items():
                    self.logger.info("\t- %s" % key_import)
                    if not self.install_mode() and import_data.get('on_install_only', False):
                        self.logger.info("\t\t* skipped not install mode")
                        return
                    func = self.get_func(import_data)
                    func(import_data.get('file_path', ''), import_data.get('model', ''), params=import_data)



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import logging
import os.path
import sys

import hiyapyco

from .apps import connection
from .apps import account
from .apps import config
from .apps import datas
from .apps import defaults
from .apps import imports
from .apps import modules
from .apps import roles
from .apps import translations
from .apps import users
from .apps import website
from .apps import mattermost
from .apps import call
from .apps import import_configurator

from .import_manager import ImportManager
from .logging import get_logger
from .odoo_connection import OdooConnection
from .password_manager import PasswordManager

logger = get_logger(__name__)


class Configurator:
    mode = ["config"]
    debug = False
    connection = None
    import_manager = None
    config = dict()
    pre_update_config = dict()
    xmlid_cache = dict()

    def __init__(self, paths=False, install=False, update=False, debug=False, debug_xmlrpc=False, keepass='',
                 config_dict=None):
        self.password_manager = PasswordManager(keepass or os.environ.get('KEEPASS_PASSWORD', ''))
        self.configurator_dir = os.path.dirname(sys.argv[0])
        if install:
            self.mode.append('install')
        if update:
            self.mode.append('update')
        self.debug = debug
        self.debug_xmlrpc = debug_xmlrpc
        if paths:
            self.paths = paths
            self.config, self.pre_update_config = self.parse_config()
        else:
            self.config = config_dict
        self.log_history = []
        if debug:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

        odoo_params = self.get_odoo_auth_params()
        connection.OdooConnection(self).pre_config(odoo_params)
        self.connection = OdooConnection(
            odoo_params['url'],
            odoo_params['dbname'],
            odoo_params['username'],
            odoo_params['password'],
            version=self.config.get('version', False),
            http_user=odoo_params.get('http_user'),
            http_password=odoo_params.get('http_password'),
            createdb=odoo_params.get('create_db'),
            debug_xmlrpc=self.debug_xmlrpc,
        )
        self.import_manager = ImportManager(self)

    def get_odoo_auth_params(self):
        if self.config.get('auth') and self.config.get('auth').get('odoo'):
            return self.config['auth']['odoo']
        else:
            return {
                'url': os.environ.get('ODOO_URL'),
                'dbname': os.environ.get('ODOO_DB'),
                'username': os.environ.get('ODOO_USER'),
                'password': os.environ.get('ODOO_PASSWORD'),
            }

    def parse_config(self):
        count_inherit = 0
        count_pre_update = 0
        pre_update_config = None
        parsed_config = hiyapyco.load(self.paths, method=hiyapyco.METHOD_MERGE, interpolate=True,
                                      failonmissingfiles=True,
                                      loglevel='INFO')

        while len(parsed_config.get("pre_update", [])) != count_pre_update:
            count_pre_update = len(parsed_config.get("pre_update", []))
            config_files = self.get_files_path(parsed_config['pre_update'])
            logger.info("Pre Update Loading %s" % (",".join(config_files)))
            pre_update_config = hiyapyco.load(config_files, method=hiyapyco.METHOD_MERGE, interpolate=True,
                                              failonmissingfiles=True, loglevel='INFO')
            pre_update_config['auth'] = parsed_config['auth']

        while len(parsed_config.get("inherits", [])) != count_inherit:
            count_inherit = len(parsed_config.get("inherits", []))
            config_files = self.paths + self.get_files_path(parsed_config['inherits']) + self.paths
            logger.info("Configuration Loading %s" % (",".join(config_files)))
            parsed_config = hiyapyco.load(config_files, method=hiyapyco.METHOD_MERGE, interpolate=True,
                                          failonmissingfiles=True, loglevel='INFO')

        parsed_config['scripts'] = []
        count_script = 0
        while len(parsed_config.get("script_files", [])) != count_script:
            count_script = len(parsed_config.get("script_files", []))
            script_files = self.get_files_path(parsed_config['script_files'])

            for script_file in script_files:
                parsed_script = hiyapyco.load(script_file, method=hiyapyco.METHOD_MERGE, interpolate=True,
                                              failonmissingfiles=True, loglevel='INFO')
                if parsed_script.get('title'):
                    parsed_script['title'] = '%s : %s' % (os.path.basename(script_file),
                                                          parsed_script.get('title'))
                else:
                    parsed_script['title'] = os.path.basename(script_file)
                parsed_config['scripts'].append(parsed_script)

        return parsed_config, pre_update_config

    def get_files_path(self, files):
        res = []
        template_dirs = [
            os.path.join(self.configurator_dir, 'templates'),
            os.path.join(self.configurator_dir, 'src/templates'),
            os.path.join(os.path.dirname(__file__), '../templates'),
        ]
        for file in files:
            if os.path.isfile(file):
                res.append(file)
            else:
                file_found = ''
                for template_dir in template_dirs:
                    file_path = os.path.join(template_dir, file)
                    if os.path.isfile(file_path):
                        file_found = file_path
                        res.append(file_path)
                        continue
                if file_found:
                    continue

                for path in self.paths:
                    file_path = os.path.join(os.path.dirname(path), file)
                    if os.path.isfile(file_path):
                        file_found = file_path
                        continue

                if file_found:
                    res.append(file_found)
                else:
                    logger.info("File not found: %s" % file)
        return res

    def get_log(self):
        return "\n".join(self.log_history)

    def show(self):
        pass
        # logger.info('show')
        # logger.info(pformat(self.config))

    def start(self):
        translations.OdooTranslations(self)
        modules_manager = modules.OdooModules(self)
        modules_manager.pre_update_config_modules()
        datas_manager = datas.OdooDatas(self)
        datas_manager.execute_pre_update_config_datas()
        modules_manager.install_config_modules()
        datas_manager.execute_update_config_datas()
        config.OdooConfig(self)
        roles.OdooRoles(self)
        defaults.OdooDefaults(self)
        users.OdooUsers(self)
        account.OdooAccount(self)
        website.OdooWebsite(self)
        imports.OdooImports(self)
        import_configurator.ImportConfigurator(self)
        mattermost.Mattermost(self)
        call.OdooCalls(self)
        return self.get_log()



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import base64
import logging
import os.path
import sys
import pickle
import ssl
import xmlrpc.client
from pprint import pformat
from .logging import get_logger

import requests
from bs4 import BeautifulSoup

CACHE = "/tmp/.configurator_cache"

METHODE_MAPPING = {
    15: [('get_object_reference', 'check_object_reference')]
}

class OdooConnection:
    _context = {'lang': 'fr_FR', 'noupdate': True}
    _cache = {}

    def __init__(self, url, dbname, user, password, version=False, http_user=None, http_password=None, createdb=False,
                 debug_xmlrpc=False):
        self.logger = get_logger("Odoo Connection".ljust(15))
        if debug_xmlrpc:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)
        self._url = url
        self._dbname = dbname
        self._user = user
        self._password = password
        self._http_user = http_user
        self._http_password = http_password
        self._version = version
        # noinspection PyProtectedMember,PyUnresolvedReferences
        self._insecure_context = ssl._create_unverified_context()
        self._load_cache()
        self._compute_url()
        if createdb:
            self._create_db()
        self._prepare_connection()

    @property
    def context(self):
        return self._context

    def _load_cache(self):
        self._cache = {}
        if os.path.isfile(CACHE):
            with open(CACHE, "rb") as f:
                self._cache = pickle.loads(f.read())

    def _save_cache(self):
        with open(CACHE, "wb") as f:
            f.write(pickle.dumps(self._cache))

    def _compute_url(self):
        if self._http_user or self._http_password:
            self._url = self._url.replace('https://', 'https://%s:%s@' % (self._http_user, self._http_password))

    def _get_xmlrpc_method(self, method):
        new_method = method
        for v in METHODE_MAPPING:
            if self._version >= v:
                for i in METHODE_MAPPING[v]:
                    if i[0] == method:
                        new_method = i[1]
        return new_method

    def _create_db(self):
        post = {
            'master_pwd': "admin123",
            'name': self._dbname,
            'login': self._user,
            'password': self._password,
            'phone': '',
            'lang': 'fr_FR',
            'country_code': 'fr',
        }
        session = requests.Session()
        session.verify = False
        r = session.post(url=self._url + "/web/database/create", params=post)
        soup = BeautifulSoup(r.text, 'html.parser')
        alert = soup.find('div', attrs={"class": u"alert alert-danger"})
        if alert:
            self.logger.debug(self._url + "/web/database/create")
            self.logger.debug(post)
            self.logger.debug(alert.get_text())
            if "already exists" not in alert.text:
                raise Exception(alert.text)

    def _prepare_connection(self):
        self.logger.info("Prepare connection %s %s %s" % (self._url, self._dbname, self._user))
        self.common = xmlrpc.client.ServerProxy('{}/xmlrpc/2/common'.format(self._url), allow_none=True,
                                                context=self._insecure_context)
        self.object = xmlrpc.client.ServerProxy('{}/xmlrpc/2/object'.format(self._url), allow_none=True,
                                                context=self._insecure_context)
        self.uid = self.common.authenticate(self._dbname, self._user, self._password, {})
        self.logger.debug('Connection uid : %s' % self.uid)
        if not self.uid:
            raise Exception("Connection Error to %s %s %s" % (self._url, self._dbname, self._user))

    def execute_odoo(self, *args, no_raise=False, retry=0):
        self.logger.debug("*" * 50)
        self.logger.debug("Execute odoo :")
        self.logger.debug("\t Model : %s" % (args[0]))
        self.logger.debug("\t Method : %s" % (args[1]))
        self.logger.debug("\t " + "%s " * (len(args) - 2) % args[2:])
        self.logger.debug("*" * 50)
        try:
            res = self.object.execute_kw(self._dbname, self.uid, self._password, *args)
            return res
        except Exception as e:
            if no_raise:
                return
            if retry <= 3:
                return self.execute_odoo(*args, retry=retry+1)
            else:
                self.logger.error(pformat(args))
                if isinstance(e, xmlrpc.client.Fault):
                    self.logger.error(e.faultString, exc_info=True)
                else:
                    self.logger.error(e)
                raise e

    def get_ref(self, external_id):
        res = self.execute_odoo('ir.model.data',
                                self._get_xmlrpc_method('get_object_reference'),
                                external_id.split('.'))[1]
        self.logger.debug('Get ref %s > %s' % (external_id, res))
        return res

    def get_image_url(self, url):
        if 'image_url' not in self._cache:
            self._cache['image_url'] = {}
        if url not in self._cache['image_url']:
            self._cache['image_url'][url] = base64.b64encode(requests.get(url).content).decode("utf-8", "ignore")
            self._save_cache()
        return self._cache['image_url'][url]

    def get_image_local(self, path):
        if 'path' not in self._cache:
            self._cache['path'] = {}
        if not os.path.isfile(path):
            path = os.path.join(os.path.dirname(sys.argv[1]), 'datas', path)

        if path not in self._cache['path']:

            self._cache['path'][path] = base64.b64encode(open(path, "rb").read()).decode("utf-8", "ignore")
            self._save_cache()
        return self._cache['path'][path]

    @staticmethod
    def get_local_file(path, encode=False):
        if encode:
            with open(path, "rb") as f:
                res = f.read()
            res = base64.b64encode(res).decode("utf-8", "ignore")
        else:
            with open(path, "r") as f:
                res = f.read()
        return res

    def get_country(self, code):
        return self.execute_odoo('res.country', 'search', [[('code', '=', code)], 0, 1, "id", False],
                                 {'context': self._context})[0]

    def get_menu(self, website_id, url):
        return self.execute_odoo('website.menu', 'search',
                                 [[('website_id', '=', website_id), ('url', '=', url)], 0, 1, "id", False],
                                 {'context': self._context})[0]

    def get_search_id(self, model, domain, order='asc'):
        res = self.execute_odoo(model, 'search', [domain, 0, 1, "id %s" % order, False], {'context': self._context})
        return res[0] if res else False

    def get_id_from_xml_id(self, xml_id, no_raise=False):
        if '.' not in xml_id:
            xml_id = "external_config." + xml_id
        try:
            object_reference = self._get_xmlrpc_method('get_object_reference')
            res_object_reference = self.execute_odoo('ir.model.data', object_reference, xml_id.split('.'),
                                                     {'context': {'active_test': False}}, no_raise=True)
            return res_object_reference[1] if res_object_reference else False
        except Exception as e:  # TODO : get true exception type and return False
            if no_raise:
                pass
            else:
                raise e

    def get_xml_id_from_id(self, model, res_id):
        try:
            domain = [('model', '=', model), ('res_id', '=', res_id)]
            res = self.execute_odoo('ir.model.data', 'search_read', [domain, ['module', 'name'], 0, 0, "id"],
                                    {'context': self._context})[0]
            return "%s.%s" % (res['module'], res['name'])
        except Exception as e:  # TODO : get true exception type and return False
            raise e
            # return False

    def set_active(self, is_active, model, domain, search_value_xml_id):
        if search_value_xml_id:
            object_id = self.get_id_from_xml_id(search_value_xml_id)
            domain = [(domain[0][0], domain[0][1], object_id)]
        object_ids = self.execute_odoo(model, 'search', [domain, 0, 0, "id", False], {'context': self._context})
        self.execute_odoo(model, 'write', [object_ids, {'active': is_active}], {'context': self._context})

    def read_search(self, model, domain, context=False):
        res = self.execute_odoo(model, 'search_read',
                                [domain],
                                {'context': context or self._context})
        return res



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import csv
import os
import sys
from .logging import get_logger
from datetime import datetime
import unidecode
import re
from pprint import pformat

class ImportManager:

    def __init__(self, configurator):
        self.logger = get_logger("Imports ".ljust(15))
        self._connection = configurator.connection
        self._context_base = self._connection.context.copy()
        self._context_base.update({'tracking_disable': True, '__import__': True})
        self._context = self._context_base

        self.name_create_enabled_fields = ''
        self.context = False
        self.limit = 0
        self.skip_line = 0
        self.batch_size = 1000
        self.thread = 0
        self.ignore_errors = []

    @staticmethod
    def parse_csv_file(file_path, delimiter=","):
        vals = []
        with open(os.path.dirname(__file__) + '/' + file_path, 'r') as csvfile:
            reader = csv.reader(csvfile, skipinitialspace=True, delimiter=delimiter)
            next(reader)
            for line in reader:
                vals.append(line)
        return vals

    @staticmethod
    def clean_field(field):
        if "/" in field:
            field = field.split("/")[0]
        if "." in field:
            field = field.split(".")[0]
        return field

    @staticmethod
    def field_check_integer(field, data):
        if field['name'] == 'id':
            return data
        if not data:
            return data
        data = int(data)
        return data

    @staticmethod
    def field_check_float(field, data):
        if not data:
            return data
        if isinstance(data, str) and "," in data:
            data = data.replace(',', '.')
        data = float(data)
        return data

    @staticmethod
    def field_check_monetary(field, data):
        return ImportManager.field_check_float(field, data)

    @staticmethod
    def field_check_date(field, data):
        if not data:
            return data
        format_ok = False
        try:
            datetime.strptime(data, '%Y-%m-%d')
            format_ok = True
        except:
            format_ok = False
        if not format_ok:
            try:
                data = datetime.strptime(data, '%d/%m/%Y').strftime('%Y-%m-%d')
                format_ok = True
            except Exception as e:
                format_ok = False
        return data

    @staticmethod
    def field_check_datetime(field, data):
        if not data:
            return data
        format_ok = False
        try:
            datetime.strptime(data, '%Y-%m-%d %H:%M:%S')
            format_ok = True
        except Exception as e:
            format_ok = False
        if not format_ok:
            try:
                data = datetime.strptime(data, '%d/%m/%Y %H:%M:%S').strftime('%Y-%m-%d %H:%M:%S')
                format_ok = True
            except Exception as e:
                format_ok = False
        if not format_ok:
            data = ImportManager.field_check_date(field, data)
            data += " 00:00:00"

        return data

    # @staticmethod
    def parse_csv_file_dictreader(self, file_path, fields, delimiter=","):
        vals = []
        new_file_path = ''
        if not os.path.isfile(file_path):
            new_file_path = os.path.dirname(__file__) + '/' + file_path
        if not os.path.isfile(new_file_path):
            new_file_path = os.path.join(os.path.dirname(sys.argv[1]), 'datas', file_path)
        if os.path.isfile(new_file_path):
            file_path = new_file_path
        else:
            print("ERROR: file %s not found." % file_path)
            return False


        with open(file_path, 'r') as csvfile:
            reader = csv.DictReader(csvfile, skipinitialspace=True, delimiter=delimiter, quotechar='"')
            for i in range(self.skip_line):
                next(reader)
            cc = 0
            for line in reader:
                if self.limit and cc == self.limit:
                    break
                cc += 1
                for field in line:
                    clean_field = ImportManager.clean_field(field)
                    method = "field_check_%s" % (fields.get(clean_field,{}).get('ttype'))
                    if hasattr(ImportManager, method):
                        method_to_call = getattr(ImportManager, method)
                        line[field] = method_to_call(fields[clean_field], line[field])
                vals.append(line)


        return vals

    @staticmethod
    def name_to_xml_id(name):
        unaccented_string = unidecode.unidecode(name)
        return re.sub('[^A-Za-z0-9]+', '', unaccented_string).replace(' ', '_').lower()

    def load_batch(self, model, datas):
        if not datas:
            return
        cc_max = len(datas)
        start = datetime.now()

        load_keys = list(datas[0].keys())
        load_datas = [[]]
        cc = 0
        for data in datas:
            if len(load_datas[-1]) >= self.batch_size:
                load_datas.append([])
            cc += 1
            # print(load_keys)
            # print(data)
            load_datas[-1].append([data[i] for i in load_keys])

        cc = 0
        for load_data in load_datas:
            start_batch = datetime.now()
            self.logger.info("\t\t* %s : %s-%s/%s" % (model, self.skip_line + cc, self.skip_line + cc + len(load_data), self.skip_line + cc_max))
            cc += len(load_data)
            res = self._connection.execute_odoo(model, 'load', [load_keys, load_data], {'context': self._context})
            for message in res['messages']:
                if message.get('message'):
                    to_ignore = [ignore_error for ignore_error in self.ignore_errors if ignore_error in message['message']]
                else:
                    to_ignore = []
                if message.get('type') in ['warning', 'error']:
                    if not to_ignore:
                        if message.get('record'):
                            self.logger.error("record : %s" % (message['record']))
                        if message.get('message'):
                            self.logger.error("message : %s" % (message['message']))
                        # for l in load_data:
                        #     print("----")
                        #     for i in range(len(load_keys)):
                        #         print(i, load_keys[i],l[i])
                        raise Exception(message['message'])
                else:
                    self.logger.info(message)
            stop_batch = datetime.now()
            self.logger.info("\t\t\tBatch time %s ( %sms per object)" % (
                stop_batch - start_batch, ((stop_batch - start_batch) / len(load_data)).microseconds / 1000))

        stop = datetime.now()
        self.logger.info("\t\t\tTotal time %s" % (stop - start))

    def set_params(self, params):

        if not params:
            params = {
                'name_create_enabled_fields': "",
                'batch_size': 1000,
                'limit': 0,
                'skip_line': 0,
                'ignore_errors': [],
                'context': False}
        self.name_create_enabled_fields = params.get('name_create_enabled_fields')
        self.context = params.get('context', False)
        self.limit = params.get('limit', 0)
        self.skip_line = params.get('skip_line', 0)
        self.batch_size = params.get('batch_size', 1000)
        self.ignore_errors = params.get('ignore_errors', [])

        self._context = self._context_base.copy()
        if self.name_create_enabled_fields:
            create_enabled_fields = dict([x, 1] for x in set(self.name_create_enabled_fields.split(",")))
            self._context.update({'name_create_enabled_fields': create_enabled_fields})
        if self.context:
            self._context.update(self.context)

    def import_csv(self, file_path, model, params=dict):
        self.set_params(params)
        fields = self.get_model_fields(model)
        raw_datas = self.parse_csv_file_dictreader(file_path, fields)
        if raw_datas:
            self.load_batch(model, raw_datas)

    def get_model_fields(self, model):
        return {i['name']: i for i in self._connection.execute_odoo("ir.model.fields", 'search_read',
                                                                    [[('model', '=', model)],
                                                                     ['name', 'ttype', 'display_name']],
                                                                    {'context': self._context})}



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from collections import OrderedDict
import requests
from ast import literal_eval
from zipfile import ZipFile
import os
from . import base


class ImportConfigurator(base.OdooModule):
    _name = "Import Configurator"

    def apply(self):
        super(ImportConfigurator, self).apply()
        for key in self._datas:
            if isinstance(self._datas.get(key), dict) or isinstance(self._datas.get(key), OrderedDict):
                action_name = 'import_configurator'
                import_configurator = self._datas.get(key).get(action_name, {})
                if import_configurator:

                    self.logger.info("\t- %s : %s" % (key, action_name))
                    res = self.execute_odoo('ir.config_parameter', 'search_read',
                                            [[('key', '=', 'odoo_configurator.access.token')],
                                             ['value']],
                                            {'context': self._context})
                    if res:
                        token = res[0].get('value', False)
                        s = requests.Session()
                        s.get(url='%s/web?db=%s' % (self._configurator.connection._url,
                                                    self._configurator.connection._dbname))

                        url = '%s/teclib_configurator_generator/get_configuration/%s' % \
                              (self._configurator.connection._url, token)
                        r = s.get(url=url)

                        open('odoo_configurator.zip', 'wb').write(r.content)
                        dest_path = os.path.dirname(self._configurator.paths[0])+'/config'
                        with ZipFile('odoo_configurator.zip', 'r') as zip:
                            zip_files = zip.namelist()
                            for export_file in import_configurator:
                                for zip_file in zip_files:
                                    if zip_file.endswith('/'+export_file) and '/modules/' in zip_file:
                                        file_bin = zip.read(zip_file)
                                        open(dest_path+'/'+export_file, 'wb').write(file_bin)
                                        self.logger.info("\t\t--> %s Updated" % export_file)

                #######################
                #  CONFIGURATOR FILE  #
                #######################
                action_name = 'import_configurator_model_file'
                configurator_model_files = self._datas.get(key).get(action_name, {})
                if configurator_model_files:

                    self.logger.info("\t- %s : %s" % (key, action_name))
                    res = self.execute_odoo('ir.config_parameter', 'search_read',
                                            [[('key', '=', 'odoo_configurator.access.token')],
                                             ['value']],
                                            {'context': self._context})
                    if res:
                        token = res[0].get('value', False)
                        s = requests.Session()
                        s.get(url='%s/web?db=%s' % (self._configurator.connection._url,
                                                    self._configurator.connection._dbname))

                        for configurator_model_file in configurator_model_files:
                            model_file = literal_eval(configurator_model_file)
                            if isinstance(model_file, str):
                                model = model_file
                                domain = '[]'
                                fields = '[]'
                            else:
                                model = model_file[0]
                                domain = str(model_file[1]) if len(model_file) > 1 else '[]'
                                fields = str(model_file[2]) if len(model_file) > 2 else '[]'
                            name = model.replace('.', '_')
                            dest_path = os.path.dirname(self._configurator.paths[0]) + '/config'
                            url = '%s/teclib_configurator_generator/get_configuration_file/%s' % \
                                  (self._configurator.connection._url, token)
                            r = s.get(url=url, params={'domain': domain, 'model': model, 'fields': fields})
                            open('%s/%s.yml' % (dest_path, name), 'wb').write(r.content)

                ##############################
                #  CONFIGURATOR BINARY FILE  #
                ##############################
                action_name = 'import_configurator_binary_file'
                configurator_model_files = self._datas.get(key).get(action_name, {})
                if configurator_model_files:
                    res = self.execute_odoo('ir.config_parameter', 'search_read',
                                            [[('key', '=', 'odoo_configurator.access.token')],
                                             ['value']],
                                            {'context': self._context})
                    if res:
                        token = res[0].get('value', False)
                    else:
                        return
                    url = '%s/teclib_configurator_generator/get_binary_file/%s' % \
                          (self._configurator.connection._url, token)
                    self.logger.info("\t- %s : %s" % (key, action_name))
                    res = self.execute_odoo('ir.config_parameter', 'search_read',
                                            [[('key', '=', 'odoo_configurator.access.token')],
                                             ['value']],
                                            {'context': self._context})
                    if res:
                        token = res[0].get('value', False)
                        s = requests.Session()
                        s.get(url='%s/web?db=%s' % (self._configurator.connection._url,
                                                    self._configurator.connection._dbname))

                        for configurator_model_file in configurator_model_files:
                            model_file = literal_eval(configurator_model_file)
                            model = model_file[0]
                            field = model_file[1]
                            res_ids = model_file[2]
                            for res_id in res_ids:
                                dest_path = os.path.dirname(self._configurator.paths[0]) + '/datas'

                                r = s.get(url=url, params={'model': model, 'field': field, 'res_id': res_id})

                                if r.headers['content-disposition']:
                                    file_name = r.headers['content-disposition'].split('=')[1].split("''")[1]
                                else:
                                    file_name = '%s_%s.bin' % (model.replace('.', '_'), res_id)
                                open('%s/%s' % (dest_path, file_name), 'wb').write(r.content)



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from collections import OrderedDict

from . import base


class OdooRoles(base.OdooModule):
    _name = "Roles"

    def apply(self):
        super(OdooRoles, self).apply()

        for key in self._datas:
            if isinstance(self._datas.get(key), dict) or isinstance(self._datas.get(key), OrderedDict):
                for key_import, data in self._datas.get(key, {}).get('datas_roles', {}).items():
                    self.logger.info("\t- %s" % key_import)
                    force_id = data.get('force_id', False)
                    role_name = data['values'].get("name")
                    role_id = self.execute_odoo('res.users.role', 'search',
                                                [[('name', '=', role_name)], 0, 0, "id", False],
                                                {'context': self._context})
                    implied_ids = [(5,), ]
                    for group in data['values'].get("implied_ids"):
                        implied_ids.append(
                            (4, self._connection.get_ref(group)))

                    user_ids = []
                    for user in data['values'].get("line_ids", []):
                        user_ids.append(
                            (4, self._connection.get_ref(user)))

                    vals = {
                        'id': force_id,
                        'name': role_name,
                        'implied_ids': implied_ids,
                        'line_ids': user_ids,
                    }
                    if not role_id:
                        self.execute_odoo('res.users.role', 'create', [vals], {'context': self._context})
                    else:
                        self.execute_odoo('res.users.role', 'write', [role_id, vals], {'context': self._context})



# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from collections import OrderedDict
from ast import literal_eval

from . import base
from .config import OdooConfig


def prepare_load_values(load_fields, fields, values):
    load_values = []
    for load_field in load_fields:
        if load_field in fields:
            value = values[fields.index(load_field)]
        else:
            value = False
        load_values.append(value)

    return load_values


class OdooDatas(base.OdooModule):
    _name = "Datas"

    def apply(self):
        pass  # For remove standard log

    def execute(self, datas):
        for key in datas:
            if isinstance(datas.get(key), dict) or isinstance(datas.get(key), OrderedDict):
                data = datas.get(key).get('datas', {})
                if data:
                    self.logger.info("\t- %s" % key)
                    self.odoo_datas(data)

        scripts = datas.get('scripts', [])
        odoo_config = OdooConfig(self._configurator)
        for script in scripts:
            self.logger.info("Script - %s" % script.get('title'))
            odoo_config.execute_script_config(script)
            self.execute(script)

    def execute_pre_update_config_datas(self):
        self.logger.info("Apply Pre-%s" % self._name)
        if self._pre_datas:
            self.execute(self._pre_datas)

    def execute_update_config_datas(self):
        self.logger.info("Apply %s" % self._name)
        self.get_external_config_xmlid_cache()
        self.execute(self._datas)

    def get_external_config_xmlid_cache(self):
        domain = [['module', '=', 'external_config']]
        datas = self.execute_odoo('ir.model.data', 'search_read', [domain, ['name', 'res_id']], {'context': {}})
        for data in datas:
            self._configurator.xmlid_cache['external_config.%s' % data['name']] = data['res_id']

    def execute_config(self, config):
        if config:
            self.pre_config(config)
            self._connection.execute_config(config)

    def odoo_datas(self, datas):
        self.pre_config(datas)
        raw_load_values = []
        load_fields = []
        load = datas.pop('load', False)
        model = datas.pop('model', False)
        for data in datas:
            object_ids = False
            self.logger.info("\t\t* %s" % data)
            if not isinstance(datas[data], dict):
                return
            if 'install' not in self._mode and datas[data].get('on_install_only', False):
                return
            model = datas[data].get('model') or model
            key = datas[data].get('key', False)
            force_id = datas[data].get('force_id', False)
            delete_all = datas[data].get('delete_all', False)
            delete_domain = datas[data].get('delete_domain', False)
            delete_id = datas[data].get('delete_id', False)
            deactivate = datas[data].get('deactivate', False)
            activate = datas[data].get('activate', False)
            update_domain = datas[data].get('update_domain', False)
            search_value_xml_id = datas[data].get('search_value_xml_id', False)
            action_server_id = datas[data].get('action_server_id', False)
            function = datas[data].get('function', False)
            no_raise = datas[data].get('no_raise', False)
            context = dict(datas[data].get('context', {}))
            values = datas[data].get('values', {})
            languages = datas[data].get('languages', [self._context['lang']])

            # escape {{
            for val_key in values:
                val = values[val_key]
                if isinstance(val, str) and '\\{\\{' in val:
                    values[val_key] = val.replace('\\{\\{', '{{')

            config_context = self._context.copy()
            config_context.update(context)

            if delete_all or delete_domain:
                if 'install' not in self._mode and datas[data].get('on_install_only', False):
                    continue
                if delete_domain:
                    domain = self.eval_param_value(delete_domain)
                else:
                    domain = []
                object_ids = self.execute_odoo(model, 'search', [domain, 0, 0, "id", False],
                                               {'context': config_context})
                for object_id in object_ids:
                    self.execute_odoo(model, 'unlink', [object_id])
                continue

            if delete_id:
                try:
                    object_id = self._connection.get_id_from_xml_id(delete_id, no_raise=True)
                    if object_id:
                        self.execute_odoo(model, 'unlink', [object_id])
                except Exception as e:
                    self.logger.error(e)
                    pass
                continue

            if update_domain:
                domain = self.eval_param_value(update_domain)
                if search_value_xml_id:
                    object_id = self._connection.get_id_from_xml_id(search_value_xml_id)
                    domain = [(domain[0][0], domain[0][1], object_id)]
                object_ids = self.execute_odoo(model, 'search', [domain, 0, 0, "id", False],
                                               {'context': config_context})
                self.logger.debug("Update Domain %s %s" % (len(object_ids), model))
                self.execute_odoo(model, 'write', [object_ids, dict(values)], {'context': config_context})
                continue

            if deactivate:
                self._connection.set_active(False, model, literal_eval(deactivate), search_value_xml_id)
                continue

            if activate:
                self._connection.set_active(True, model, literal_eval(activate), search_value_xml_id)
                continue

            if key:
                object_ids = self.execute_odoo(model, 'search',
                                               [[(key, '=', values[key])], 0, 0, "id", False],
                                               {'context': config_context})
            elif force_id:
                if load:
                    values['id'] = force_id
                else:
                    values, object_ids = self.set_force_id(model, values, config_context, force_id)
            elif action_server_id:
                self.exec_action_server(action_server_id, datas[data], config_context, no_raise)
                continue
            elif function:
                self.exec_function(function, datas[data], config_context, no_raise)
                continue
            else:
                self.logger.error("key=%s", key)
                raise

            # prepare many2many list of xmlid
            for key in values.keys():
                if isinstance(values[key], list) and '/id' in key:
                    if values[key] and isinstance(values[key][0], str) and '.' in values[key][0]:
                        values[key] = ','.join(values[key])

            if load:

                fields, rec_values = self.save_values(model, values, config_context, force_id, object_ids,
                                                      load_batch=load)
                load_fields = list(set(load_fields + fields))
                raw_load_values.append((fields, rec_values[0]))
            else:
                for language in languages:
                    config_context['lang'] = language
                    self.save_values(model, values, config_context, force_id, object_ids)

        if load:
            load_values = [prepare_load_values(load_fields, v[0], v[1]) for v in raw_load_values]
            context = self._context
            self.execute_load(model, load_fields, load_values, context)

    def save_values(self, model, values, config_context, force_id, object_ids, load_batch=False):
        if not force_id or isinstance(force_id, int):
            if object_ids:
                self.execute_odoo(model, 'write', [object_ids, dict(values)], {'context': config_context})
            else:
                self.execute_odoo(model, 'create', [dict(values)], {'context': config_context})
        else:
            load_keys = list(values.keys())
            load_data = [[]]
            for i in load_keys:
                if isinstance(values[i], bool):
                    load_data[0].append(str(values[i]))
                else:
                    load_data[0].append(values[i])
            self.logger.debug('%s: %s %s' % (model, load_keys, load_data))
            if load_batch:
                return load_keys, load_data
            else:
                self.execute_load(model, load_keys, load_data, config_context)

    def execute_load(self, model, load_keys, load_data, config_context):
        res = self.execute_odoo(model, 'load', [load_keys, load_data], {'context': config_context})
        for message in res['messages']:
            self.logger.error("%s : %s" % (message['record'], message['message']))

    def set_force_id(self, model, values, config_context, force_id=False):
        if isinstance(force_id, int):
            values['id'] = force_id
            object_ids = self.execute_odoo(model, 'search', [[('id', '=', force_id)], 0, 0, "id", False],
                                           {'context': config_context})
        else:
            if '.' not in force_id:
                force_id = "external_config." + force_id
            values['id'] = force_id
            if force_id in self._configurator.xmlid_cache:
                return values, self._configurator.xmlid_cache[force_id]
            module, name = force_id.split('.')
            if self.execute_odoo('ir.model.data', 'search',
                                 [[('module', '=', module), ('name', '=', name)], 0, 0, "id", False],
                                 {'context': config_context}):
                object_ids = self._connection.get_ref(force_id)
            else:
                object_ids = []
        return values, object_ids

    def exec_action_server(self, action_server_id, datas, context, no_raise):
        context.update({'active_model': datas.get('model'),
                        'active_id': datas.get('res_id')})
        res = self.execute_odoo('ir.actions.server', 'run', [action_server_id], {'context': context},
                                no_raise=no_raise)

    def eval_param_value(self, param):
        if isinstance(param, str):
            if param.startswith('get_'):
                return self.safe_eval(param)
            if param[0] in ['[', '{', '(']:
                param_val = literal_eval(param)
                return param_val

        if isinstance(param, OrderedDict):
            return dict(param)
        if isinstance(param, list):
            param_val = []
            for param_el in param:
                param_el_val = self.eval_param_value(param_el)
                param_val.append(param_el_val)
            return param_val

        return param

    def exec_function(self, function, datas, context, no_raise):
        params = datas.get('params')
        if params:
            res_id = datas.get('res_id', False)
            function_params = [res_id] if res_id else [0]
            for param in params:
                param_val = self.eval_param_value(param)
                function_params.append(param_val)

            extra_params = {'context': context}
            kw = datas.get('kw')
            if kw:
                extra_params.update(dict(kw))

            res = self.execute_odoo(datas.get('model'), function, function_params, extra_params,
                                    no_raise=no_raise)
        else:
            res = self.execute_odoo(datas.get('model'), function, [datas.get('res_id')], {'context': context},
                                    no_raise=no_raise)



from . import base
from .modules import OdooModules
import requests
import json
import os

# -*- coding: utf-8 -*-
# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

class Mattermost(base.OdooModule):
    _name = "Mattermost"
    _key = "mattermost"

    def apply(self):
        super(Mattermost, self).apply()
        if self._datas.get('no_notification', False):
            return

        mattermost_channel = self._datas.get('mattermost_channel', False)
        mattermost_url = self._datas.get('mattermost_url', False)

        if not mattermost_channel or not mattermost_url:
            return

        headers = {}
        values = {
            "username": "Configurator : %s" % (os.getlogin()),
            "channel": mattermost_channel,
            "text": """#### Odoo %s updated with Configurator
- Url : %s
- Db : %s
- User : %s""" % (self._datas.get('name', "NONAME"),
                  self._datas.get('auth', {}).get('odoo', {}).get('url'),
                  self._datas.get('auth', {}).get('odoo', {}).get('dbname'),
                  self._datas.get('auth', {}).get('odoo', {}).get('username'),
                  )
        }
        if "localhost" not in self._datas.get('auth', {}).get('odoo', {}).get('url'):
            response = requests.post(mattermost_url,
                                     headers=headers, data=json.dumps(values))
        else:
            print("No send notification on localhost")
